from __future__ import unicode_literals

import venusian


def consumer_config(*args, **kwargs):
    """Decorator for configure the given consumer function. For example,
    you want to consume a certain type of events from the queue, here you can
    write

        @consumer_config(
            topic='justitia.generic_events',
            cls_types=('metrics', ),
        )
        def process_metrics(event):
            '''Process metrics events and push values to librato

            '''
            # the code goes here

    """
    def callback(scanner, name, ob):
        consumer_cls = kwargs.pop('consumer_cls', Consumer)
        if 'name' not in kwargs:
            kwargs['name'] = name
        consumer = consumer_cls(ob, *args, **kwargs)
        scanner.add_consumer(consumer)

    def decorator(wrapped):
        venusian.attach(wrapped, callback, category='balog.consumers')
        return wrapped

    return decorator


class Consumer(object):
    """A consumer represents a function with certain configurations for
    consuming events

    """

    def __init__(self, func, topic, cls_types=None, name=None):
        self.func = func
        self.topic = topic
        self.cls_types = cls_types
        self.name = name

    def __repr__(self):
        return '<{} {}>'.format(
            self.__class__.__name__,
            self.name or self.func,
        )

    def match_event(self, event):
        """Return whether is this consumer interested in the given event

        """
        if self.cls_types is None:
            return True
        if event['payload']['cls_type'] in self.cls_types:
            return True
        return False


class ConsumerHub(object):
    """Consumer hub is a collection of consumers, you can feed it with events,
    and it knows which consumers to route those events to

    """

    def __init__(self):
        self.consumers = []

    def add_consumer(self, consumer):
        """Add a function as a consumer

        """
        self.consumers.append(consumer)

    def scan(self, package, ignore=None):
        """Scan packages for finding consumers

        """
        scanner = venusian.Scanner(add_consumer=self.add_consumer)
        scanner.scan(
            package,
            categories=('balog.consumers', ),
            ignore=ignore,
        )

    def route(self, event):
        """Return a list of consumers for the given event should be routed to

        """
        for consumer in self.consumers:
            if not consumer.match_event(event):
                continue
            yield consumer



from __future__ import unicode_literals
import datetime
import logging

import pytz
import colander

from balog.records.application import ApplicationRecordSchema
from balog.guid import GUIDFactory

LOG_GUID_FACTORY = GUIDFactory('LG')
logger = logging.getLogger(__name__)


@colander.deferred
def deferred_utcnow(node, kw):
    return datetime.datetime.utcnow().replace(tzinfo=pytz.utc)


@colander.deferred
def deferred_guid(node, kw):
    return LOG_GUID_FACTORY()


class EventContext(colander.MappingSchema):
    fqdn = colander.SchemaNode(colander.String(), default=colander.drop)
    application = colander.SchemaNode(colander.String(), default=colander.drop)
    application_version = colander.SchemaNode(colander.String(), default=colander.drop)


class EventHeader(colander.MappingSchema):
    id_ = colander.SchemaNode(
        colander.String(),
        name='id',
        default=deferred_guid,
    )
    channel = colander.SchemaNode(colander.String())
    timestamp = colander.SchemaNode(colander.DateTime(), default=deferred_utcnow)
    composition = colander.SchemaNode(colander.Bool(), default=colander.drop)
    context = EventContext(default=colander.drop)


class FacilityRecordSchema(colander.MappingSchema):
    VERSION = '0.0.1'

    schema = colander.SchemaNode(colander.String(), default=VERSION)
    header = EventHeader()
    payload = ApplicationRecordSchema()



from __future__ import unicode_literals
import uuid

import coid


class GUIDFactory(object):
    """Object for making prefixed GUID

    """

    def __init__(self, prefix):
        self.prefix = prefix
        self.id_codec = coid.Id(prefix=self.prefix, encoding='base58')

    def __call__(self):
        return self.id_codec.encode(uuid.uuid4())



from __future__ import unicode_literals
import unittest

from balog.consumers import ConsumerHub
from ..fixtures import my_consumers


class TestConsumer(unittest.TestCase):

    def test_consumer_hub(self):
        hub = ConsumerHub()
        hub.scan(my_consumers)

        self.assertEqual(len(hub.consumers), 3)
        func_to_consumers = dict(
            (consumer.func, consumer)
            for consumer in hub.consumers
        )
        self.assertEqual(
            set(func_to_consumers.keys()),
            set(
                [
                    my_consumers.consumer_a,
                    my_consumers.consumer_b,
                    my_consumers.consumer_c
                ]
            )
        )

        consumer_a = func_to_consumers[my_consumers.consumer_a]
        self.assertEqual(consumer_a.topic, 'foo.bar')
        self.assertEqual(consumer_a.cls_types, ('eggs', ))

        consumer_b = func_to_consumers[my_consumers.consumer_b]
        self.assertEqual(consumer_b.topic, 'spam.eggs')
        self.assertEqual(consumer_b.cls_types, None)

        consumer_c = func_to_consumers[my_consumers.consumer_c]
        self.assertEqual(consumer_c.topic, '5566')
        self.assertEqual(consumer_c.cls_types, ('55', '66'))

        eggs_event = dict(payload=dict(cls_type='eggs'))
        self.assertEqual(
            set(hub.route(eggs_event)),
            set([consumer_a, consumer_b]),
        )



from __future__ import unicode_literals
import json
import collections
import functools
import threading
import logging
import time

import boto.sqs
from boto.sqs.message import RawMessage

from ..records.facility import FacilityRecordSchema

# Notice: this logger should be configured carefully
logger = logging.getLogger(__name__)


class DefaultConsumerOperator(object):

    def get_topic(self, consumer):
        """Get the topic of consumer

        """
        return consumer.topic

    def process_event(self, consumer, event):
        """Call consumer's function

        """
        return consumer.func(event)


# TODO: define a base class?
class SQSEngine(object):
    """Event processing engine for polling Amazon SQS queues

    """

    def __init__(
        self,
        hub,
        region,
        aws_access_key_id,
        aws_secret_access_key,
        polling_period=1,
        num_messages=10,
        consumer_operator=None,
    ):
        self.hub = hub
        # aws credentials
        self.region = region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        #: polling period in seconds
        self.polling_period = polling_period
        #: maximum number of messages to get at once
        self.num_messages = num_messages
        #: the object which knows how to operate consumers
        self.consumer_operator = consumer_operator
        if self.consumer_operator is None:
            self.consumer_operator = DefaultConsumerOperator

        self.running = False

    def _poll_topic(self, topic, consumers):
        """Poll events from SQS

        """
        logger.info(
            'Polling %s for consumers %s',
            topic, consumers,
        )
        schema = FacilityRecordSchema()
        queue = self.conn.get_queue(topic)
        queue.set_message_class(RawMessage)
        while self.running:
            msgs = queue.get_messages(
                num_messages=self.num_messages,
                wait_time_seconds=self.polling_period,
            )
            for msg in msgs:
                json_data = json.loads(msg.get_body())
                event = schema.deserialize(json_data)
                logger.debug('Processing event %r', event)
                # Notice: Since we're processing logs, if we generate
                # log and that will be consumed by this loop, it may
                # end up with flood issue (one log generates more logs)
                # so, we should be careful, do not generate log record
                # from this script
                for consumer in consumers:
                    self.consumer_operator.process_event(consumer, event)
                # delete it from queue
                queue.delete_message(msg)

    def run(self):
        self.running = True
        self.conn = boto.sqs.connect_to_region(
            self.region,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )
        # map topic name (queue name) to consumers
        topic_to_consumers = collections.defaultdict(list)
        for consumer in self.hub.consumers:
            topic = self.consumer_operator.get_topic(consumer)
            topic_to_consumers[topic].append(consumer)

        # create threads for consuming events from SQS
        threads = []
        for topic, consumers in topic_to_consumers.iteritems():
            thread = threading.Thread(
                target=functools.partial(self._poll_topic, topic, consumers),
                name='polling-topic-{}-worker'.format(topic)
            )
            thread.daemon = True
            threads.append(thread)

        # start all threads
        for thread in threads:
            thread.start()

        try:
            while self.running:
                time.sleep(1)
        except (SystemExit, KeyboardInterrupt):
            self.running = False
            logger.info('Stopping SQS engine')

        for thread in threads:
            thread.join()
        logger.info('Stopped SQS engine')



from ez_setup import use_setuptools
use_setuptools()

from setuptools import setup, find_packages

import os

tests_require = [
    'nose',
    'nose-cov',
    'mock',
    'jsonschema',
]

sqs_require = [
    'boto',
]

version = open(os.path.join('balog', 'version.txt'), 'rt').read()


setup(
    name='balog',
    version=version,
    description='Balanced event logging schema and library',
    url='https://github.com/balanced/balog/',
    packages=find_packages(),
    install_requires=[
        'jsonschema',
        'structlog',
        'pytz',
        'iso8601',
        'coid',
        'venusian',
        # Notice: the original package name is colander, we have a branch here
        # git+git@github.com:victorlin/colander.git@polymorphism#egg=colander
        # since we cannot add the git repo in install_requires, so I release
        # my branch as ba-colander. I know there is `dependency_links`,
        # however, it's not recommended by pip, for more information, you
        # can reference to
        # https://groups.google.com/forum/#!topic/pypa-dev/tJ6HHPQpyJ4
        'ba-colander==1.0b1',
    ],
    extras_require=dict(
        tests=tests_require,
        sqs=sqs_require,
    ),
    tests_require=tests_require,
    test_suite='nose.collector',
)



from __future__ import unicode_literals
import os

import structlog
from structlog.stdlib import LoggerFactory
from structlog.processors import format_exc_info
from structlog.processors import JSONRenderer

from .processors import LogProcessor


def load_pkg_file(pkg_filename, filename, default):
    """Load file content under package folder

    """
    pkg_dir = os.path.dirname(pkg_filename)
    filepath = os.path.join(pkg_dir, filename)
    try:
        with open(filepath, 'rt') as pkg_file:
            return pkg_file.read().strip()
    except IOError:
        return default

__version__ = load_pkg_file(__file__, 'version.txt', '0.0.0')


def configure():
    """Configure Balanced logging system

    """
    structlog.configure(
        processors=[
            format_exc_info,
            LogProcessor(),
            JSONRenderer(),
        ],
        logger_factory=LoggerFactory(),
    )


def get_logger(*args, **kwargs):
    """Get and return the logger

    """
    return structlog.get_logger(*args, **kwargs)



from __future__ import unicode_literals

from balog.consumers import consumer_config


@consumer_config(topic='foo.bar', cls_types=('eggs', ))
def consumer_a(event):
    pass


@consumer_config(topic='spam.eggs')
def consumer_b(event):
    pass


@consumer_config(topic='5566', cls_types=('55', '66'))
def consumer_c(event):
    pass



from __future__ import unicode_literals
import logging

import colander
from colander.polymorphism import AbstractSchema


logger = logging.getLogger(__name__)


class ApplicationRecordSchema(AbstractSchema):
    cls_type = colander.SchemaNode(colander.String())

    # TODO: this is what I think how it should work
    # need to implement a MappingSchema so that it can work like this
    __mapper_args__ = {
        'polymorphic_on': 'cls_type',
    }


class Log(ApplicationRecordSchema):
    cls_type = 'log'
    message = colander.SchemaNode(colander.String())
    severity = colander.SchemaNode(colander.String(), validators=[
        colander.OneOf((
            'debug',
            'info',
            'warning',
            'error',
            'critical',
        ))
    ])
    exc_text = colander.SchemaNode(colander.String(), default=colander.drop)


class MetricsValue(colander.MappingSchema):
    name = colander.SchemaNode(colander.String())
    value = colander.SchemaNode(colander.Float())


class MetricsValues(colander.SequenceSchema):
    value = MetricsValue()


class Metrics(ApplicationRecordSchema):
    cls_type = 'metrics'
    values = MetricsValues()



from __future__ import unicode_literals
import json

from .records.facility import FacilityRecordSchema


class LogProcessor(object):

    @classmethod
    def jsonify_unstructed_log(cls, record):
        """Jsonify a Python logging library record

        """
        channel = record.name
        schema = FacilityRecordSchema()
        event_dict = schema.bind().serialize({
            'header': {
                'channel': channel,
                'context': {
                    # TODO: add application information?
                }
            },
            'payload': {
                'cls_type': 'log',
                'severity': record.levelname.lower(),
                'message': record.getMessage(),
            }
        })
        if record.exc_text:
            event_dict['payload']['exc_text'] = record.exc_text
        return json.dumps(event_dict)

    def __call__(self, logger, name, event_dict):
        """Transform given event dict to the format we want

        """
        channel = event_dict.pop('channel', None)
        if channel is None:
            channel = logger.name + '.' + event_dict.pop('event')
        schema = FacilityRecordSchema()
        new_event_dict = schema.bind().serialize({
            'header': {
                'channel': channel,
                'context': {
                    # TODO: add application information?
                }
            },
            'payload': event_dict['payload']
        })
        return new_event_dict



from __future__ import unicode_literals
import logging
import json

from .processors import LogProcessor


class SchemaFormatter(logging.Formatter):
    """Logging formatter for making sure record are turned into defined json
    schema

    """

    def format(self, record):
        msg = super(SchemaFormatter, self).format(record)
        try:
            json.loads(msg)
        # looks like we have a plain text string message
        except ValueError:
            # transform it into log schema
            msg = LogProcessor.jsonify_unstructed_log(record)
        return msg
